import os
import sys
import time

from mesos.interface import ExecutorDriver, mesos_pb2

from .messages_pb2 import RegisterExecutorMessage, \
    ExecutorToFrameworkMessage, StatusUpdateMessage
from .process import UPID, Process, async


class MesosExecutorDriver(Process, ExecutorDriver):
    def __init__(self, executor):
        Process.__init__(self, 'executor')
        self.executor = executor

        env = os.environ
        self.local = bool(env.get('MESOS_LOCAL'))
        slave_pid = env.get('MESOS_SLAVE_PID')
        assert slave_pid, 'expecting MESOS_SLAVE_PID in environment'
        self.slave = UPID(slave_pid)
        self.framework_id = mesos_pb2.FrameworkID()
        self.framework_id.value = env.get('MESOS_FRAMEWORK_ID')
        self.executor_id = mesos_pb2.ExecutorID()
        self.executor_id.value = env.get('MESOS_EXECUTOR_ID')
        self.workDirectory = env.get('MESOS_DIRECTORY')

    def onExecutorRegisteredMessage(self, executor_info, framework_id,
            framework_info, slave_id, slave_info):
        assert framework_id == self.framework_id
        self.slave_id = slave_id
        return self.executor.registered(self, executor_info, framework_info,
            slave_info)

    def onRunTaskMessage(self, framework_id, framework, pid, task):
        return self.executor.launchTask(self, task)

    def onKillTaskMessage(self, framework_id, task_id):
        return self.executor.killTask(self, task_id)

    def onFrameworkToExecutorMessage(self, slave_id, framework_id,
            executor_id, data):
        return self.executor.frameworkMessage(self, data)

    def onShutdownExecutorMessage(self):
        self.executor.shutdown(self)
        if not self.local:
            sys.exit(0)
        else:
            self.stop()

    def onStatusUpdateAcknowledgementMessage(self, slave_id, framework_id, task_id, uuid):
        pass

    def start(self):
        Process.start(self)
        msg = RegisterExecutorMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        msg.executor_id.MergeFrom(self.executor_id)
        return self.send(self.slave, msg)

    @async
    def sendFrameworkMessage(self, data):
        msg = ExecutorToFrameworkMessage()
        msg.slave_id.MergeFrom(self.slave_id)
        msg.framework_id.MergeFrom(self.framework_id)
        msg.executor_id.MergeFrom(self.executor_id)
        msg.data = data
        return self.send(self.slave, msg)

    @async
    def sendStatusUpdate(self, status):
        msg = StatusUpdateMessage()
        msg.update.framework_id.MergeFrom(self.framework_id)
        msg.update.executor_id.MergeFrom(self.executor_id)
        msg.update.slave_id.MergeFrom(self.slave_id)
        msg.update.status.MergeFrom(status)
        msg.update.timestamp = time.time()
        msg.update.uuid = os.urandom(16)
        return self.send(self.slave, msg)



try:
    from zookeeper import ZooKeeperException as ZookeeperError
    from zkpython import ZKClient, ChildrenWatch, DataWatch
    def adjust_zk_logging_level():
        pass
except ImportError:
    from kazoo.client import KazooClient as ZKClient
    from kazoo.recipe.watchers import ChildrenWatch, DataWatch
    from kazoo.exceptions import ZookeeperError
    def adjust_zk_logging_level():
        import logging
        import kazoo
        kazoo.client.log.setLevel(logging.WARNING)
        kazoo.protocol.connection.log.setLevel(logging.WARNING)

class MasterDetector(object):
    def __init__(self, uri, agent):
        self.uri = uri
        self.agent = agent
        self.zk = ZKClient(uri, 10)
        self.masterSeq = None

    def choose(self, children):
        if not children:
            self.agent.onNoMasterDetectedMessage()
            return True
        masterSeq = min(children)
        if masterSeq == self.masterSeq:
            return True
        self.masterSeq = masterSeq
        DataWatch(self.zk, '/' + masterSeq, self.notify)
        return True

    def notify(self, master_addr, _):
        self.agent.onNewMasterDetectedMessage(master_addr)
        return False

    def start(self):
        adjust_zk_logging_level()
        self.zk.start()
        try:
            ChildrenWatch(self.zk, '', self.choose)
        except ZookeeperError:
            self.agent.onNoMasterDetectedMessage()
            self.stop()

    def stop(self):
        try: self.zk.stop()
        except: pass


def test():
    import logging
    logging.basicConfig()
    class Agent:
        def onNewMasterDetectedMessage(self, addr):
            print 'got', addr
        def onNoMasterDetectedMessage(self):
            print 'no master'
    d = MasterDetector('zk1:2181/mesos_master2', Agent())
    d.start()
    raw_input("press any key to exit:\n")

if __name__ == '__main__':
    test()



import os, sys
import subprocess
import pwd
import logging

logger = logging.getLogger(__name__)

class Launcher(object):
    """
    This class sets up the environment for an executor and then exec()'s it.
    It can either be used after a fork() in the slave process, or run as a
    standalone program (with the main function in launcher_main.cpp).

    The environment is initialized through for steps:
        1) A work directory for the framework is created by createWorkingDirectory().
        2) The executor is fetched off HDFS if necessary by fetchExecutor().
        3) Environment variables are set by setupEnvironment().
        4) We switch to the framework's user in switchUser().

        Isolation modules that wish to override the default behaviour can subclass
        Launcher and override some of the methods to perform extra actions.
    """
    def __init__(self, framework_id, executor_id, commandInfo, user=None, workDirectory='/tmp',
            slavepid=None, redirectIO=False, switch_user=False):
        self.__dict__.update(locals())

    def run(self):
        self.initializeWorkingDirectory()
        os.chdir(self.workDirectory)
        if self.switch_user:
            self.switchUser()
        if self.redirectIO:
            sys.stdout = open('stdout', 'w')
            sys.stderr = open('stderr', 'w')
        command = self.commandInfo.value
        env = self.setupEnvironment()
        p = subprocess.Popen(['/bin/sh', '-c', command], stdout=sys.stdout,
                stderr=sys.stderr, env=env)
        p.wait()

    def setupEnvironment(self):
        env = {}
        for en in self.commandInfo.environment.variables:
            env[en.name] = en.value
        env['MESOS_DIRECTORY'] = self.workDirectory
        env['MESOS_SLAVE_PID'] = str(self.slavepid)
        env['MESOS_FRAMEWORK_ID'] = self.framework_id.value
        env['MESOS_EXECUTOR_ID'] = self.executor_id.value
        env['LIBPROCESS_PORT'] = '0'
        return env

    def initializeWorkingDirectory(self):
        if self.switch_user:
            try:
                os.chown(self.user, self.workDirectory)
            except IOError, e:
                logger.error("failed to chown: %s", e)

    def switchUser(self):
        try:
            pw = pwd.getpwnam(self.user)
            os.setuid(pw.pw_uid)
            os.setgid(pw.pw_gid)
        except OSError, e:
            logger.error("failed to swith to user %s: %s", self.user, e)

def main():
    from mesos.interace.mesos_pb2 import FrameworkID, ExecutorID, CommandInfo
    fid = FrameworkID()
    fid.value = os.environ.get('MESOS_FRAMEWORK_ID', 'fid')
    eid = ExecutorID()
    eid.value = os.environ.get('MESOS_EXECUTOR_ID', 'eid')
    info = CommandInfo()
    info.value = os.environ.get('MESOS_COMMAND')
    return Launcher(fid, eid, info).run()

if __name__ == '__main__':
    main()



import os, sys
import time
import logging
import socket
import threading
import signal

from mesos.interace.mesos_pb2 import *

from .messages_pb2 import *
from .process import UPID, Process
from .launcher import Launcher


logger = logging.getLogger(__name__)

class Resources(object):
    def __init__(self, cpus, mem):
        self.cpus = cpus
        self.mem = mem

    def __iadd__(self, other):
        if not isinstance(other, Resources):
            other = Resources.new(other)
        self.cpus += other.cpus
        self.mem += other.mem
        return self

    def __isub__(self, other):
        if not isinstance(other, Resources):
            other = Resources.new(other)
        self.cpus -= other.cpus
        self.mem -= other.mem
        return self

    @classmethod
    def new(cls, rs):
        self = cls(0, 0)
        for r in rs:
            self.__dict__[r.name] = r.scalar.value
        return self


class Executor(object):
    def __init__(self, framework_id, info, directory):
        self.framework_id = framework_id
        self.info = info
        self.directory = directory

        self.id = info.executor_id
        self.resources = Resources.new(info.resources)
        self.uuid = os.urandom(16)

        self.pid = None
        self.shutdown = False
        self.launchedTasks = {}
        self.queuedTasks = {}

    def addTask(self, task):
        assert task.task_id.value not in self.launchedTasks
        t = Task()
        t.framework_id.MergeFrom(self.framework_id)
        t.state = TASK_STAGING
        t.name = task.name
        t.task_id.MergeFrom(task.task_id)
        t.slave_id.MergeFrom(task.slave_id)
        t.resources.MergeFrom(task.resources)
        if task.command:
            t.executor_id.MergeFrom(self.id)

        self.launchedTasks[task.task_id.value] = t
        self.resources += task.resources
        return t

    def removeTask(self, task_id):
        tid = task_id.value
        self.queuedTasks.pop(tid, None)
        if tid in self.launchedTasks:
            self.resources -= self.launchedTasks[tid].resources
            del self.launchedTasks[tid]

    def updateTaskState(self, task_id, state):
        if task_id.value in self.launchedTasks:
            self.launchedTasks[task_id.value].state = state


class Framework(object):
    def __init__(self, id, info, pid, conf):
        self.id = id
        self.info = info
        self.pid = pid
        self.conf = conf

        self.executors = {}
        self.updates = {}

    def get_executor_info(self, task):
        assert task.HasField('executor') != task.HasField('command')
        if task.HasField('executor'):
            return task.executor

        info = ExecutorInfo()
        fmt = "Task %s (%s)"
        if len(task.command.value) > 15:
            id = fmt % (task.task_id.value, task.command.value[:12] + '...')
        else:
            id = fmt % (task.task_id.value, task.command.value)
        info.executor_id.value = id
        info.framework_id.value = self.id
        info.command.value = task.command.value
        return info

    def createExecutor(self, info, directory):
        executor = Executor(self.id, info, directory)
        self.executors[executor.id.value] = executor
        return executor

    def destroyExecutor(self, exec_id):
        self.executors.pop(exec_id.value, None)

    def get_executor(self, exec_id):
        if isinstance(exec_id, TaskID):
            tid = exec_id.value
            for executor in self.executors.values():
                if tid in executor.queuedTasks or tid in executor.launchedTasks:
                    return executor
            return
        return self.executors.get(exec_id.value)
    getExecutor = get_executor


class ProcessInfo(object):
    def __init__(self, fid, eid, pid, directory):
        self.framework_id = fid
        self.executor_id = eid
        self.pid = pid
        self.directory = directory


class IsolationModule(object):
    def __init__(self):
        self.infos = {}
        self.pids = {}

    def initialize(self, slave):
        self.slave = slave
        t = threading.Thread(target=self.reaper)
        t.daemon = True
        t.start()

    def launchExecutor(self, framework_id, info, executor_info, directory, resources):
        fid = framework_id.value
        eid = executor_info.executor_id.value
        logger.info("Launching %s (%s) in %s with resources %s for framework %s",
                eid, executor_info.command.value, directory, resources, fid)
        pid = os.fork()
        assert pid >= 0
        if pid:
            logger.info("Forked executor at %d", pid)
            pinfo = ProcessInfo(framework_id, executor_info.executor_id, pid, directory)
            self.infos.setdefault(fid, {})[eid] = pinfo
            self.pids[pid] = pinfo
            self.slave.executorStarted(framework_id, executor_info.executor_id, pid)
        else:
            #pid = os.setsid()
            # copy UPID of slave
            launcher = Launcher(framework_id, executor_info.executor_id, executor_info.command,
                    info.user, directory, self.slave)
            launcher.run()

    def killExecutor(self, framework_id, executor_id):
        pinfo = self.infos.get(framework_id.value, {}).get(executor_id.value)
        if pinfo is None:
            logger.error("ERROR! Asked to kill an unknown executor! %s", executor_id.value)
            return
        pid = pinfo.pid
        os.kill(pid, signal.SIGKILL)
        # remove when waitpid()
        #self._removeExecutor(pinfo)

    def removeExecutor(self, e):
        logger.info("remove executor: %s", e)
        if e.framework_id.value not in self.infos:
            return
        self.infos[e.framework_id.value].pop(e.executor_id.value, None)
        if not self.infos[e.framework_id.value]:
            del self.infos[e.framework_id.value]
        self.pids.pop(e.pid)

    def resourcesChanged(self, framework_id, executor_id, resources):
        pass

    def reaper(self):
        while True:
            if not self.pids:
                time.sleep(1)
                continue

            try:
                pid, status = os.waitpid(-1, 0)
            except OSError, e:
                time.sleep(1)
                logger.error("waitpid: %s", e)
                continue

            if pid > 0 and pid in self.pids and not os.WIFSTOPPED(status):
                e = self.pids[pid]
                logger.info("Telling slave of lost executor %s of framework %s",
                        e.executor_id, e.framework_id)
                self.slave.executorExited(e.framework_id, e.executor_id, status)
                self.removeExecutor(e)


def isTerminalTaskState(state):
    return state in (TASK_FINISHED, TASK_FAILED, TASK_KILLED, TASK_LOST)


class Slave(Process):
    def __init__(self, options):
        Process.__init__(self, "slave")
        self.options = options
        self.resources = Resources(options.cpus, options.mem)
        self.attributes = options.attributes

        self.id = None
        self.isolation = IsolationModule()
        self.info = self.getSlaveInfo()
        self.master = UPID('master', options.master)
        self.frameworks = {}
        self.startTime = time.time()
        self.connected = False

    def getSlaveInfo(self):
        info = SlaveInfo()
        info.hostname = socket.gethostname()
        info.webui_hostname = ''
        info.webui_port = 8081
        cpus = info.resources.add()
        cpus.name = 'cpus'
        cpus.type = 0
        cpus.scalar.value = self.resources.cpus
        mem = info.resources.add()
        mem.name = 'mem'
        mem.type = 0
        mem.scalar.value = self.resources.mem
        if self.attributes:
            for attrs in self.attributes.split(','):
                name,value = attrs.split(':')
                a = self.info.attributes.add()
                a.name = name
                a.type = Value.TEXT # 3 # TEXT
                a.text.value = value
        return info

    def getFramework(self, framework_id):
        return self.frameworks.get(framework_id.value)

    def onNewMasterDetectedMessage(self, pid):
        self.master = UPID(pid)
        self.register()

    def onNoMasterDetectedMessage(self):
        self.master = None
        self.connected = False

    def register(self):
        if self.id is None:
            msg = RegisterSlaveMessage()
            msg.slave.MergeFrom(self.info)
        else:
            msg = ReregisterSlaveMessage()
            msg.slave_id.MergeFrom(self.id)
            msg.slave.MergeFrom(self.info)
            for framework in self.frameworks.itervalues():
                for executor in framework.executors.values():
                    msg.executor_infos.add().MergeFrom(executor.info)
                    for task in executor.launchedTasks.itervalues():
                        msg.tasks.add().MergeFrom(task)
        return self.send(self.master, msg)

    def onSlaveRegisteredMessage(self, slave_id):
        logger.info("slave registed %s", slave_id.value)
        self.id = slave_id
        self.connected = True

    def onSlaveReregisteredMessage(self, slave_id):
        assert self.id == slave_id
        self.connected = True

    def onRunTaskMessage(self, framework_id, framework_info, pid, task):
        logger.info("Got assigned task %s for framework %s",
                task.task_id.value, framework_id.value)
        fid = framework_id.value
        if fid not in self.frameworks:
            framework = Framework(framework_id, framework_info, UPID(pid), self.options)
            self.frameworks[fid] = framework
        else:
            framework = self.frameworks[fid]

        executorInfo = framework.get_executor_info(task)
        eid = executorInfo.executor_id
        executor = framework.getExecutor(eid)
        if executor:
            if executor.shutdown:
                logger.warning("WARNING! executor is shuting down")
            elif not executor.pid:
                executor.queuedTasks[task.task_id.value] = task
            else:
                executor.addTask(task)
                #self.isolation.resourcesChanged(framework_id, executor.id, executor.resources)

                msg = RunTaskMessage()
                msg.framework.MergeFrom(framework.info)
                msg.framework_id.MergeFrom(framework_id)
                msg.pid = str(framework.pid)
                msg.task.MergeFrom(task)
                self.send(executor.pid, msg)
        else:
            directory = self.createUniqueWorkDirectory(framework.id, eid)
            executor = framework.createExecutor(executorInfo, directory)
            executor.queuedTasks[task.task_id.value] = task
            self.isolation.launchExecutor(framework.id, framework.info, executor.info,
                    directory, executor.resources)

    def onKillTaskMessage(self, framework_id, task_id):
        framework = self.getFramework(framework_id)
        if not framework or not framework.getExecutor(task_id):
            msg = StatusUpdateMessage()
            update = msg.update
            update.framework_id.MergeFrom(framework_id)
            update.slave_id.MergeFrom(self.id)
            update.status.task_id.MergeFrom(task_id)
            update.status.state = TASK_LOST
            update.timestamp = time.time()
            update.uuid = os.urandom(16)
            return self.send(self.master, msg)

        executor = framework.getExecutor(task_id)
        if not executor.pid:
            executor.removeTask(task_id)
            msg = StatusUpdateMessage()
            update = msg.update
            update.framework_id.MergeFrom(framework_id)
            update.slave_id.MergeFrom(self.id)
            update.status.task_id.MergeFrom(task_id)
            update.status.state = TASK_KILLED
            update.timestamp = time.time()
            update.uuid = os.urandom(16)
            return self.send(self.master, msg)

        msg = KillTaskMessage()
        msg.framework_id.MergeFrom(framework_id)
        msg.task_id.MergeFrom(task_id)
        return self.send(executor.pid, msg)

    def onShutdownFrameworkMessage(self, framework_id):
        framework = self.getFramework(framework_id)
        if framework:
            for executor in framework.executors.values():
                self.shutdownExecutor(framework, executor)

    def shutdownExecutor(self, framework, executor):
        logger.info("shutdown %s %s", framework.id.value, executor.id.value)
        self.send(executor.pid, ShutdownExecutorMessage())
        executor.shutdown = True

        # delay check TODO
        time.sleep(3)
        self.isolation.killExecutor(framework.id, executor.id)
        framework.destroyExecutor(executor.id)
        if not framework.executors and not framework.updates:
            self.frameworks.pop(framework.id.value)

    def onUpdateFrameworkMessage(self, framework_id, pid):
        framework = self.getFramework(framework_id)
        if framework:
            framework.pid = pid

    def executorStarted(self, framework_id, executor_id, pid):
        pass

    def executorExited(self, framework_id, executor_id, status):
        logger.info("Executor %s of framework %s exited with %d",
                executor_id.value, framework_id.value, status)
        framework = self.getFramework(framework_id)
        if not framework:
            return
        executor = framework.getExecutor(executor_id)
        if not executor:
            return

        isCommandExecutor = False
        for task in executor.launchedTasks.values():
            if not isTerminalTaskState(task.state):
                isCommandExecutor = not task.HasField('executor_id')
                self.transitionLiveTask(task.task_id, executor_id,
                    framework_id, isCommandExecutor, status)

        for task in executor.queuedTasks.values():
            isCommandExecutor = task.HasField('command')
            self.transitionLiveTask(task.task_id, executor_id,
                    framework_id, isCommandExecutor, status)

        if not isCommandExecutor:
            msg = ExitedExecutorMessage()
            msg.slave_id.MergeFrom(self.id)
            msg.framework_id.MergeFrom(framework_id)
            msg.executor_id.MergeFrom(executor_id)
            msg.status = status
            self.send(self.master, msg)
        framework.destroyExecutor(executor_id)

    def onRegisterExecutorMessage(self, framework_id, executor_id):
        framework = self.getFramework(framework_id)
        if not framework:
            # TODO shutdown executor
            return
        executor = framework.getExecutor(executor_id)
        if not executor or executor.pid or executor.shutdown:
            # TODO shutdown executor
            return
        executor.pid = self.sender

        msg = ExecutorRegisteredMessage()
        msg.executor_info.MergeFrom(executor.info)
        msg.framework_id.MergeFrom(framework_id)
        msg.framework_info.MergeFrom(framework.info)
        msg.slave_id.MergeFrom(self.id)
        msg.slave_info.MergeFrom(self.info)
        self.send(executor.pid, msg)

        for task in executor.queuedTasks.values():
            msg = RunTaskMessage()
            msg.framework_id.MergeFrom(framework.id)
            msg.framework.MergeFrom(framework.info)
            msg.pid = str(framework.pid)
            msg.task.MergeFrom(task)
            self.send(executor.pid, msg)

        for task in executor.queuedTasks.values():
            executor.addTask(task)
        executor.queuedTasks.clear()

    def onStatusUpdateMessage(self, update):
        status = update.status
        framework = self.getFramework(update.framework_id)
        if not framework:
            return
        executor = framework.getExecutor(status.task_id)
        if not executor:
            return
        executor.updateTaskState(status.task_id, status.state)
        if isTerminalTaskState(status.state):
            executor.removeTask(status.task_id)

        msg = StatusUpdateMessage()
        msg.update.MergeFrom(update)
        msg.pid = str(self) # pid
        self.send(self.master, msg)

        framework.updates[update.uuid] = update

    def onStatusUpdateAcknowledgementMessage(self, slave_id, framework_id, task_id, uuid):
        framework = self.getFramework(framework_id)
        if framework and uuid in framework.updates:
            framework.updates.pop(uuid)
            if not framework.executors and not framework.updates:
                self.frameworks.pop(framework_id.value)

    def onFrameworkToExecutorMessage(self, slave_id, framework_id, executor_id, data):
        framework = self.getFramework(framework_id)
        if not framework:
            return
        executor = framework.getExecutor(executor_id)
        if not executor:
            return
        if not executor.pid:
            return
        msg = FrameworkToExecutorMessage()
        msg.slave_id.MergeFrom(slave_id)
        msg.framework_id.MergeFrom(framework_id)
        msg.executor_id.MergeFrom(executor_id)
        msg.data = data
        self.send(executor.pid, msg)

    def onExecutorToFrameworkMessage(self, slave_id, framework_id, executor_id, data):
        framework = self.getFramework(framework_id)
        if not framework:
            return
        msg = ExecutorToFrameworkMessage()
        msg.slave_id.MergeFrom(slave_id)
        msg.framework_id.MergeFrom(framework_id)
        msg.executor_id.MergeFrom(executor_id)
        msg.data = data
        self.send(framework_id.pid, msg)

    def onShutdownMessage(self):
        self.stop()

    def start(self):
        Process.start(self)
        self.isolation.initialize(self)
        self.register() # master detector TODO

    def onPing(self):
        self.reply("PONG")

    def createStatusUpdate(self, task_id, executor_id, framework_id,
            taskState, reason):
        status = TaskStatus()
        status.task_id.MergeFrom(task_id)
        status.state = taskState
        status.message = reason

        update = StatusUpdate()
        update.framework_id.MergeFrom(framework_id)
        update.slave_id.MergeFrom(self.id)
        update.executor_id.MergeFrom(executor_id)
        update.status.MergeFrom(status)
        update.timestamp = time.time()
        update.uuid = os.urandom(16)

        return update

    def statusUpdate(self, update):
        logger.info("status update")
        status = update.status
        framework = self.getFramework(update.framework_id)
        if not framework:
            return
        executor = framework.getExecutor(status.task_id)
        if not executor:
            return
        if isTerminalTaskState(status.state):
            executor.removeTask(status.task_id)
        msg = StatusUpdateMessage()
        msg.update.MergeFrom(update)
        msg.pid = str(self)
        self.send(self.master, msg)
        # check
        framework.updates[update.uuid] = update

    def transitionLiveTask(self, task_id, executor_id, framework_id,
            isCommandExecutor, status):
        if isCommandExecutor:
            update = self.createStatusUpdate(task_id, executor_id,
                    framework_id, TASK_FAILED, "Executor running the task's command failed")
        else:
            update = self.createStatusUpdate(task_id, executor_id,
                    framework_id, TASK_LOST, "Executor exited")
        self.statusUpdate(update)

    def createUniqueWorkDirectory(self, framework_id, executor_id):
        root = self.options.work_dir
        path = os.path.join(root, 'slaves', self.id.value,
                'frameworks', framework_id.value,
                'executors', executor_id.value, 'runs')
        for i in range(10000):
            p = os.path.join(path, str(i))
            if not os.path.exists(p):
                os.makedirs(p)
                return p

def main():
    import optparse
    parser = optparse.OptionParser(usage="Usage: %prog [options]")
    parser.add_option("-s", "--master", type="string", default="localhost:5050",
            help="May be one of:\n"
              " host[:5050]"
              "  host:port\n"
              "  zk://host1:port1,host2:port2,.../path\n"
              "  zk://username:password@host1:port1,host2:port2,.../path\n"
              "  file://path/to/file (where file contains one of the above)")
    parser.add_option("-c", "--cpus", type="int", default=2)
    parser.add_option("-m", "--mem", type="int", default=1024)
    parser.add_option("-a", "--attributes", type="string")
    parser.add_option("-w", "--work_dir", type="string", default="/tmp/mesos")
    parser.add_option("-q", "--quiet", action="store_true")
    parser.add_option("-v", "--verbose", action="store_true")

    options, args = parser.parse_args()
    logging.basicConfig(format='[slave] %(asctime)-15s %(message)s',
                    level=options.quiet and logging.ERROR
                        or options.verbose and logging.DEBUG
                        or logging.WARNING)

    slave = Slave(options)
    slave.run()

if __name__ == '__main__':
    main()



import socket
import time
import threading
import logging
import Queue
import select

from mesos.interface.mesos_pb2 import *
from .messages_pb2 import *

logger = logging.getLogger(__name__)

def spawn(target, *args, **kw):
    t = threading.Thread(target=target, name=target.__name__, args=args, kwargs=kw)
    t.daemon = True
    t.start()
    return t

class UPID(object):
    def __init__(self, name, addr=None):
        if addr is None and name and '@' in name:
            name, addr = name.split('@')
        self.name = name
        self.addr = addr

    def __str__(self):
        return "%s@%s" % (self.name, self.addr)


def async(f):
    def func(self, *a, **kw):
        self.delay(0, f, self, *a, **kw)
    return func

class Process(UPID):
    def __init__(self, name, port=0):
        UPID.__init__(self, name)
        self.port = port
        self.conn_pool = {}
        self.jobs = Queue.PriorityQueue()
        self.linked = {}
        self.sender = None
        self.aborted = False

    def delay(self, delay, func, *args, **kw):
        self.jobs.put((time.time() + delay, 0, func, args, kw))

    def run_jobs(self):
        while True:
            try:
                job = self.jobs.get(timeout=1)
            except Queue.Empty:
                if self.aborted:
                    break
                continue

            #self.jobs.task_done()
            t, tried, func, args, kw = job
            now = time.time()
            if t > now:
                if self.aborted:
                    break
                self.jobs.put(job)
                time.sleep(min(t-now, 0.1))
                continue

            try:
                #logger.debug("run job %s", func.__name__)
                func(*args, **kw)
                #logger.debug("run job %s comeplete", func.__name__)
            except Exception, e:
                logger.error("error while call %s (tried %d times)", func, tried)
                import traceback; traceback.print_exc()
                if tried < 4:
                    self.jobs.put((t + 3 ** tried, tried + 1, func, args, kw))

    @async
    def link(self, upid, callback):
        self._get_conn(upid.addr)
        self.linked[upid.addr] = callback

    def _get_conn(self, addr):
        if addr not in self.conn_pool:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            host, port = addr.split(':')
            s.connect((host, int(port)))
            self.conn_pool[addr] = s
        return self.conn_pool[addr]

    def _encode(self, upid, msg):
        if isinstance(msg, str):
            body = ''
            uri = '/%s/%s' % (upid.name, msg)
        else:
            body = msg.SerializeToString()
            uri = '/%s/mesos.internal.%s' % (upid.name, msg.__class__.__name__)
        agent = 'libprocess/%s@%s' % (self.name, self.addr)
        msg = ['POST %s HTTP/1.0' % str(uri),
               'User-Agent: %s' % agent,
               'Connection: Keep-Alive',]
        if body:
            msg += [
               'Transfer-Encoding: chunked',
               '',
               '%x' % len(body), body,
               '0']
        msg += ['', ''] # for last \r\n\r\n
        return '\r\n'.join(msg)

    #@async
    def send(self, upid, msg):
        logger.debug("send to %s %s", upid, msg.__class__.__name__)
        data = self._encode(upid, msg)
        try:
            conn = self._get_conn(upid.addr)
            conn.send(data)
        except IOError:
            logger.warning("failed to send data to %s, retry again", upid)
            self.conn_pool.pop(upid.addr, None)
            if upid.addr in self.linked: # broken link
                callback = self.linked.pop(upid.addr)
                callback()
            raise

    def reply(self, msg):
        return self.send(self.sender, msg)

    def onPing(self):
        self.reply('PONG')

    @async
    def handle(self, msg):
        if self.aborted:
            return
        name = msg.__class__.__name__
        f = getattr(self, 'on' + name, None)
        assert f, 'should have on%s()' % name
        args = [v for (_,v) in msg.ListFields()]
        f(*args)

    def abort(self):
        self.aborted = True
        self.listen_sock.close()

    def stop(self):
        Process.abort(self)
        self.join()
        for addr in self.conn_pool:
            self.conn_pool[addr].close()
        self.conn_pool.clear()

    def join(self):
        self.delay_t.join()
        return self.accept_t.join()

    def run(self):
        self.start()
        return self.join()

    def start(self):
        self.listen_sock = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('0.0.0.0', self.port))
        if not self.port:
            port = sock.getsockname()[1]
            self.addr = '%s:%d' % (socket.gethostname(), port)
        self.accept_t = spawn(self.ioloop, sock)
        self.delay_t = spawn(self.run_jobs)

    def process_message(self, rf):
        headers = []
        while True:
            try:
                line = rf.readline()
            except IOError:
                break
            if not line or line == '\r\n':
                break
            headers.append(line)
        if not headers:
            return False # EoF

        method, uri, _ = headers[0].split(' ')
        _, process, mname = uri.split('/')
        assert process == self.name, 'unexpected messages'
        agent = headers[1].split(' ')[1].strip()
        logger.debug("incoming request: %s from %s", uri, agent)

        sender_name, addr = agent.split('@')
        self.sender = UPID(sender_name.split('/')[1], addr)

        if mname == 'PING':
            self.onPing()
            return True

        size = rf.readline()
        if size:
            size = int(size, 16)
            body = rf.read(size+2)[:-2]
            rf.read(5)  # ending
        else:
            body = ''

        sname = mname.split('.')[2]
        if sname not in globals():
            logger.error("unknown messages: %s", sname)
            return True

        try:
            msg = globals()[sname].FromString(body)
            self.handle(msg)
        except Exception, e:
            logger.error("error while processing message %s: %s", sname, e)
            import traceback; traceback.print_exc()
        return True

    def ioloop(self, sock):
        sock.listen(64)
        sfd = sock.fileno()
        conns = {sfd: sock}
        while not self.aborted:
            rlist = select.select(conns.keys(), [], [], 1)[0]
            for fd in rlist:
                if fd == sfd:
                    conn, addr = sock.accept()
                    logger.debug("accepted conn from %s", addr)
                    conns[conn.fileno()] = conn
                elif fd in conns:
                    try:
                        f = conns[fd].makefile('r')
                        while True:
                            if not self.process_message(f):
                                conns.pop(fd).close()
                                break
                            # is there any data in read buffer ?
                            if not f._rbuf.tell():
                                break
                        f.close()
                    except Exception, e:
                        import traceback; traceback.print_exc()
                        conns.pop(fd).close()

        sock.close()
    """
    def communicate(self, conn):
        rf = conn.makefile('r', 4096)
        while not self.aborted:
            cont = self.process_message(rf)
            if not cont:
                break
        rf.close()

    def ioloop(self, s):
        s.listen(1)
        conns = []
        while True:
            conn, addr = s.accept()
            logger.debug("accepted conn from %s", addr)
            conns.append(conn)
            if self.aborted:
                break

            spawn(self.communicate, conn)

        for c in conns:
            c.close()
    """



from setuptools import setup, find_packages

version = '0.0.2'

setup(
    name='pymesos',
    version=version,
    description="A pure python implementation of Mesos scheduler and executor",
    packages=find_packages(),
    install_requires=['mesos.interface'],
    platforms=['POSIX'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
    ],
    author="Zhongbo Tian",
    author_email="tianzhongbo@douban.com",
    url="https://github.com/douban/pymesos",
    download_url = 'https://github.com/douban/pymesos/archive/%s.tar.gz' % version,
)



import time
import logging
import struct
import socket

from mesos.interface.mesos_pb2 import TASK_LOST, MasterInfo

from .messages_pb2 import (
    RegisterFrameworkMessage, ReregisterFrameworkMessage,
    DeactivateFrameworkMessage, UnregisterFrameworkMessage,
    ResourceRequestMessage, ReviveOffersMessage, LaunchTasksMessage, KillTaskMessage,
    StatusUpdate, StatusUpdateAcknowledgementMessage, FrameworkToExecutorMessage,
    ReconcileTasksMessage
)
from .process import UPID, Process, async


logger = logging.getLogger(__name__)


class MesosSchedulerDriver(Process):
    def __init__(self, sched, framework, master_uri):
        Process.__init__(self, 'scheduler')
        self.sched = sched
        #self.executor_info = executor_info
        self.master_uri = master_uri
        self.framework = framework
        self.framework.failover_timeout = 100
        self.framework_id = framework.id

        self.master = None
        self.detector = None

        self.connected = False
        self.savedOffers = {}
        self.savedSlavePids = {}

    @async # called by detector
    def onNewMasterDetectedMessage(self, data):
        try:
            info = MasterInfo()
            info.ParseFromString(data)
            ip = socket.inet_ntoa(struct.pack('<I', info.ip))
            self.master = UPID('master@%s:%s' % (ip, info.port))
        except:
            self.master = UPID(data)

        self.connected = False
        self.register()

    @async # called by detector
    def onNoMasterDetectedMessage(self):
        self.connected = False
        self.master = None

    def register(self):
        if self.connected or self.aborted:
            return

        if self.master:
            if not self.framework_id.value:
                msg = RegisterFrameworkMessage()
                msg.framework.MergeFrom(self.framework)
            else:
                msg = ReregisterFrameworkMessage()
                msg.framework.MergeFrom(self.framework)
                msg.failover = True
            self.send(self.master, msg)

        self.delay(2, self.register)

    def onFrameworkRegisteredMessage(self, framework_id, master_info):
        self.framework_id = framework_id
        self.framework.id.MergeFrom(framework_id)
        self.connected = True
        self.link(self.master, self.onDisconnected)
        self.sched.registered(self, framework_id, master_info)

    def onFrameworkReregisteredMessage(self, framework_id, master_info):
        assert self.framework_id == framework_id
        self.connected = True
        self.link(self.master, self.onDisconnected)
        self.sched.reregistered(self, master_info)

    def onDisconnected(self):
        self.connected = False
        logger.warning("disconnected from master")
        self.delay(5, self.register)

    def onResourceOffersMessage(self, offers, pids):
        for offer, pid in zip(offers, pids):
            self.savedOffers.setdefault(offer.id.value, {})[offer.slave_id.value] = UPID(pid)
        self.sched.resourceOffers(self, list(offers))

    def onRescindResourceOfferMessage(self, offer_id):
        self.savedOffers.pop(offer_id.value, None)
        self.sched.offerRescinded(self, offer_id)

    def onStatusUpdateMessage(self, update, pid=''):
        assert self.framework_id == update.framework_id

        if pid and not pid.endswith('0.0.0.0:0'):
            reply = StatusUpdateAcknowledgementMessage()
            reply.framework_id.MergeFrom(self.framework_id)
            reply.slave_id.MergeFrom(update.slave_id)
            reply.task_id.MergeFrom(update.status.task_id)
            reply.uuid = update.uuid
            try: self.send(UPID(pid), reply)
            except IOError: pass

        self.sched.statusUpdate(self, update.status)

    def onLostSlaveMessage(self, slave_id):
        self.sched.slaveLost(self, slave_id)

    def onExecutorToFrameworkMessage(self, slave_id, framework_id, executor_id, data):
        self.sched.frameworkMessage(self, slave_id, executor_id, data)

    def onFrameworkErrorMessage(self, message, code=0):
        self.sched.error(self, code, message)

    def start(self):
        Process.start(self)
        uri = self.master_uri
        if uri.startswith('zk://') or uri.startswith('zoo://'):
            from .detector import MasterDetector
            self.detector = MasterDetector(uri[uri.index('://') + 3:], self)
            self.detector.start()
        else:
            if not ':' in uri:
                uri += ':5050'
            self.onNewMasterDetectedMessage('master@%s' % uri)

    def abort(self):
        if self.connected:
            msg = DeactivateFrameworkMessage()
            msg.framework_id.MergeFrom(self.framework_id)
            self.send(self.master, msg)
        Process.abort(self)

    def stop(self, failover=False):
        if self.connected and not failover:
            msg = UnregisterFrameworkMessage()
            msg.framework_id.MergeFrom(self.framework_id)
            self.send(self.master, msg)
        if self.detector:
            self.detector.stop()
        Process.stop(self)

    @async
    def requestResources(self, requests):
        if not self.connected:
            return
        msg = ResourceRequestMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        for req in requests:
            msg.requests.add().MergeFrom(req)
        self.send(self.master, msg)

    @async
    def reviveOffers(self):
        if not self.connected:
            return
        msg = ReviveOffersMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        self.send(self.master, msg)

    @async
    def reconcileTasks(self, statuses=None):
        if not self.connected:
            return
        msg = ReconcileTasksMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        if statuses is not None:
            msg.statuses = statuses
        self.send(self.master, msg)

    def launchTasks(self, offer_id, tasks, filters):
        if not self.connected or offer_id.value not in self.savedOffers:
            for task in tasks:
                update = StatusUpdate()
                update.framework_id.MergeFrom(self.framework_id)
                update.status.task_id.MergeFrom(task.task_id)
                update.status.state = TASK_LOST
                update.status.message = 'Master disconnected' if not self.connected else "invalid offer_id"
                update.timestamp = time.time()
                update.uuid = ''
                self.onStatusUpdateMessage(update)
            return

        msg = LaunchTasksMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        msg.offer_ids.add().MergeFrom(offer_id)
        msg.filters.MergeFrom(filters)
        for task in tasks:
            msg.tasks.add().MergeFrom(task)
            pid = self.savedOffers.get(offer_id.value, {}).get(task.slave_id.value)
            if pid and task.slave_id.value not in self.savedSlavePids:
                self.savedSlavePids[task.slave_id.value] = pid
        self.savedOffers.pop(offer_id.value)
        self.send(self.master, msg)

    def declineOffer(self, offer_id, filters=None):
        if not self.connected:
            return
        msg = LaunchTasksMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        msg.offer_ids.add().MergeFrom(offer_id)
        if filters:
             msg.filters.MergeFrom(filters)
        self.send(self.master, msg)

    @async
    def killTask(self, task_id):
        if not self.connected:
            return
        msg = KillTaskMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        msg.task_id.MergeFrom(task_id)
        self.send(self.master, msg)

    @async
    def sendFrameworkMessage(self, executor_id, slave_id, data):
        if not self.connected:
            return

        msg = FrameworkToExecutorMessage()
        msg.framework_id.MergeFrom(self.framework_id)
        msg.executor_id.MergeFrom(executor_id)
        msg.slave_id.MergeFrom(slave_id)
        msg.data = data

        slave = self.savedSlavePids.get(slave_id.value, self.master) # can not send to slave directly
        self.send(slave, msg)
